import shutil
from abc import ABC, abstractmethod
from io import StringIO
from pathlib import Path
import py_compile

import sh  # type: ignore

from .util import get_stderr, format_template, generate_random_string
from .exit_codes import ExitCodeEventType, ExitCodeHandler

GRADER_DIR = Path(__file__).resolve().parent


class TestCase(ABC):
    # Extra args you'd like to use during compilation
    SUBMISSION_COMPILATION_ARGUMENTS: tuple = tuple()
    COMPILATION_ARGUMENTS: tuple = tuple()
    source_suffix = ".source_suffix"
    executable_suffix = ".executable_suffix"
    path_to_helper_module: Path
    exit_code_handler: ExitCodeHandler = ExitCodeHandler()

    def __init__(
        self,
        path: Path,
        tests_dir: Path,
        timeout: int,
        filters,
        precompile_testcase=False
    ):
        self.path = path
        self.timeout = timeout
        self.filters = filters
        self.need_precompile_testcase = precompile_testcase
        output_dir = tests_dir / f"output/{path.stem}.txt"
        if output_dir.exists():
            with output_dir.open() as f:
                self.expected_output = self.format_output(f.read())
        else:
            self.expected_output = StringIO("")
        input_dir = tests_dir / f"input/{path.stem}.txt"
        if input_dir.exists():
            with input_dir.open() as f:
                self.input = StringIO(f.read().strip())
        else:
            self.input = StringIO("")

        # Only really works if test name is in snake_case
        self.name = path.stem.replace("_", " ").capitalize()

        self.prepend_test_helper()
        if precompile_testcase:
            self.precompile_testcase()

    def run(self, precompiled_submission: Path):
        """ Returns student score and message to be displayed """
        self.input.seek(0)
        try:
            test_executable = self.compile_testcase(precompiled_submission)
        except sh.ErrorReturnCode as e:
            return 0, get_stderr(self.path.parent, e, "Failed to compile")
        with StringIO() as runtime_output:
            try:
                result = test_executable(
                    _in=self.input,
                    _out=runtime_output,
                    _timeout=self.timeout,
                    _ok_code=self.exit_code_handler.allowed_exit_codes
                )
            except sh.TimeoutException:
                return 0, "Exceeded Time Limit"
            except sh.ErrorReturnCode as e:
                # http://man7.org/linux/man-pages/man7/signal.7.html
                return 0, f"Crashed due to signal {e.exit_code}"
            event = self.exit_code_handler.scan(result.exit_code)
            if event.type == ExitCodeEventType.CHECK_OUTPUT:
                if self.format_output(runtime_output.getvalue()) == self.expected_output:
                    return 100, "100/100"
                else:
                    return 0, "0/100 (Wrong answer)"
            elif event.type == ExitCodeEventType.RESULT:
                score = event.value
                message = f"{score}/100"
                if score == 0:
                    message += " (Wrong answer)"
                return score, message
            elif event.type == ExitCodeEventType.CHEAT_ATTEMPT:
                # This  means that either the student used built-in exit function himself
                # or some testcase helper is broken, or a testcase exits itself without
                # the use of helper functions.
                return 0, f"An invalid exit code '{result.exit_code}' has been supplied"
            elif event.type == ExitCodeEventType.SYSTEM_ERROR:
                # We should already handle this case in try, except block
                pass

    def make_executable_path(self, submission: Path):
        """ By combining test name and student name, it makes a unique path """
        return self.path.with_name(self.path.stem + submission.stem + self.executable_suffix)

    def format_output(self, output: str):
        output = output.lower()
        for filter_function in self.filters:
            output = "".join(filter(filter_function, output))
        return output

    @classmethod
    def precompile_submission(cls, submission: Path, current_dir: Path, source_file_name) -> Path:
        """ Copies student submission into currect_dir and either precompiles
            it and returns the path to the precompiled submission or to the
            copied submission if no precompilation is necesessary
        """
        destination = current_dir / "temp" / source_file_name
        shutil.copy(submission, destination)
        return destination

    def precompile_testcase(self):
        """ Replaces the original testcase file with its compiled version,
            thus making reading its contents as plaintext harder.
            Useful in preventing cheating.
        """
        pass

    @abstractmethod
    def compile_testcase(self, precompiled_submission: Path) -> sh.Command:
        """ Either precompiles student submission and returns the path to
            the precompiled submission or to original submission if no
            precompilation is necesessary
        """
        pass

    def prepend_test_helper(self):
        """ Prepends all of the associated test_helper code to test code """
        with open(self.path) as f, open(self.path_to_helper_module) as helper_file:
            formatted_helper_file = format_template(
                helper_file.read(),
                **self.exit_code_handler.get_formatted_exit_codes()
            )
            content = f.read()
            final_content = formatted_helper_file + "\n" + content
        with open(self.path, "w") as f:
            f.write(final_content)

    def cleanup(self):
        self.input.close()
        self.path.unlink()


class CTestCase(TestCase):
    source_suffix = ".c"
    executable_suffix = ".out"
    path_to_helper_module = GRADER_DIR / "test_helpers/test_helper.c"
    SUBMISSION_COMPILATION_ARGS = ("-Dscanf_s=scanf", "-Dmain=__student_main__")

    @classmethod
    def precompile_submission(cls, submission, current_dir, source_file_name):
        """ Links student submission without compiling it.
            It is done to speed up total compilation time
        """
        copied_submission = super().precompile_submission(submission, current_dir, submission.name)
        precompiled_submission = copied_submission.with_suffix(".o")
        sh.gcc(
            "-c", f"{copied_submission}",
            "-o", precompiled_submission,
            *cls.SUBMISSION_COMPILATION_ARGS
        )
        return precompiled_submission

    def precompile_testcase(self):
        sh.gcc(
            "-c", self.path,
            "-o", self.path.with_suffix('.o')
        )
        self.path.unlink()
        self.path = self.path.with_suffix('.o')

    def compile_testcase(self, precompiled_submission: Path) -> sh.Command:
        executable_path = self.make_executable_path(precompiled_submission)
        sh.gcc(
            "-o",
            executable_path,
            self.path,
            precompiled_submission.name
        )
        return sh.Command(executable_path)


class JavaTestCase(TestCase):
    """ Please, ask students to remove their main as it can theoretically
        generate errors (not sure how though).
        Java doesn't support precompilation yet. Read precompile_testcase()
        for more info.
    """
    source_suffix = ".java"
    executable_suffix = ""
    path_to_helper_module = GRADER_DIR / "test_helpers/TestHelper.java"
    parallel_execution_supported = False

    @classmethod
    def precompile_submission(cls, submission: Path, current_dir: Path, source_file_name: str):
        """ Renames submission to allow javac compilation """
        copied_submission = super().precompile_submission(submission, current_dir, submission.name)
        precompiled_submission = copied_submission.parent / source_file_name
        copied_submission.rename(precompiled_submission)
        return precompiled_submission

    def precompile_testcase(self):
        """ Sadly, java doesn't allow to compile files before their dependencies
            have compiled.
            Instead, we hide test names. It is not super safe but better than nothing.
            We could, instead, encrypt each test file and decrypt at compilation.
        """
        self.original_path = self.path
        new_name = generate_random_string(15)
        new_path = self.path.with_name(new_name)
        self.path.rename(new_path)
        self.path = self.path.with_name(new_name)

        # This method didn't work but might prove useful in the future
        # sh.javac(self.path, "-implicit:none")
        # self.path.unlink()

    def compile_testcase(self, precompiled_submission: Path) -> sh.Command:
        if self.need_precompile_testcase:
            self.path.rename(self.original_path)
            path = self.original_path
        else:
            path = self.path
        sh.javac(path, precompiled_submission.name)
        if self.need_precompile_testcase:
            path.rename(self.path)
        return lambda *args, **kwargs: sh.java(path.stem, *args, **kwargs)

    def prepend_test_helper(self):
        """ Puts private TestHelper at the end of testcase class.
            This is quite a crude way to do it but it is the easiest
            I found so far.
        """
        with open(self.path) as f, open(self.path_to_helper_module) as helper_file:
            formatted_helper_class = format_template(
                helper_file.read(),
                **self.exit_code_handler.get_formatted_exit_codes()
            )
            content = f.read()
            final_content = self._add_at_the_end_of_public_class(formatted_helper_class, content)
        with open(self.path, "w") as f:
            f.write(final_content)

    def _add_at_the_end_of_public_class(self, helper_class: str, java_file: str):
        # TODO: Figure out a better way to do this
        main_class_index = java_file.find("public")
        file_starting_from_main_class = java_file[main_class_index:]
        closing_brace_index = self._find_closing_brace(file_starting_from_main_class)
        return ''.join([
            java_file[:main_class_index],
            file_starting_from_main_class[:closing_brace_index],
            "\n" + helper_class + "\n" + "}",
            java_file[closing_brace_index + 1:]
        ])

    def _find_closing_brace(self, s: str):
        bracecount = 0
        for i in range(len(s)):
            if s[i] == "{":
                bracecount += 1
            elif s[i] == "}":
                bracecount -= 1
                if bracecount == 0:
                    return i
        else:
            raise ValueError(f"Braces in testcase '{self.name}' don't match.")


class PythonTestCase(TestCase):
    """ A proof of concept of how easy it is to add new languages.
        Will only work if python is accessible via python3 alias for now.
    """
    source_suffix = ".py"
    executable_suffix = source_suffix
    path_to_helper_module = GRADER_DIR / "test_helpers/test_helper.py"

    def compile_testcase(self, precompiled_submission: Path) -> sh.Command:
        return lambda *args, **kwargs: sh.python3(
            self.path,
            precompiled_submission.stem,
            *args,
            **kwargs
        )

    def precompile_testcase(self):
        py_compile.compile(file=self.path, cfile=self.path)



import os
import sys
import shutil
from pathlib import Path
from typing import List
import logging
import configparser

import sh  # type: ignore

from . import testcases
from .util import get_stderr
from .filters import ALLOWED_FILTERS

DEFAULT_SOURCE_FILE_STEM = "Homework"
PATH_TO_DEFAULT_CONFIG = Path(__file__).parent / "default_config.ini"
ALLOWED_LANGUAGES = {
    "c": testcases.CTestCase,
    "java": testcases.JavaTestCase,
    "python": testcases.PythonTestCase
}
# Constants
KEY = """
\nKey:
\tFailed to Compile: Your submission did not compile due to a syntax or naming error
\tCompiled with warnings: Your submission uses unchecked or unsafe operations
\tCrashed due to signal SIGNAL_CODE: Your submission threw an uncaught exception.
\tAll signal error codes are described here: http://man7.org/linux/man-pages/man7/signal.7.html
\tExceeded Time Limit: Your submission took too much time to run (probably an infinite loop)
"""


class Grader:
    def __init__(
        self,
        current_dir,
        generate_results=False,
        testcase_dir_name="testcases",
        precompile_testcases=False
    ):
        self.generate_results = generate_results
        self.current_dir = current_dir
        self.temp_dir = current_dir / "temp"
        self.tests_dir = current_dir / "tests"
        self.results_dir = current_dir / "results"
        self.path_to_output_summary = current_dir / "grader_output.txt"
        self.precompile_testcases = precompile_testcases
        self._choose_directory_structure(testcase_dir_name)
        self.temp_dir.mkdir(exist_ok=True)
        if generate_results:
            self.results_dir.mkdir(exist_ok=True)
        self._configure_grading()
        self._configure_logging()
        self.tests = self._gather_testcases()
        self.submissions = self._gather_submissions()

    def run(self):
        old_dir = Path.cwd()
        os.chdir(self.temp_dir)
        total_class_points = sum(map(self._run_tests_on_submission, self.submissions))
        class_average = total_class_points / (len(self.submissions) or 1)
        self.logger.info(f"\nAverage score: {round(class_average)}/{self.total_points_possible}")
        self.logger.info(KEY)
        self.cleanup()
        os.chdir(old_dir)
        return class_average

    def cleanup(self):
        shutil.rmtree(self.temp_dir)

    def _choose_directory_structure(self, testcase_dir_name):
        if not (self.tests_dir / testcase_dir_name).exists():
            if (self.current_dir / testcase_dir_name).exists():
                self.tests_dir = self.current_dir
        self.testcases_dir = self.tests_dir / testcase_dir_name
        REQUIRED_DIRS = (self.testcases_dir,)
        dir_not_found = "{} directory not found. It is required for the grader to function."
        for directory in REQUIRED_DIRS:
            if not directory.exists():
                raise FileNotFoundError(dir_not_found.format(directory))

    def _configure_grading(self):
        cfg = self._read_config()
        self.timeout = cfg.getint('TIMEOUT')
        self.total_points_possible = cfg.getint('TOTAL_POINTS_POSSIBLE')
        self.total_score_to_100_ratio = self.total_points_possible / 100
        language = cfg['PROGRAMMING_LANGUAGE']
        if language == "AUTO":
            self.TestCaseType = self._figure_out_testcase_type()
        else:
            self.TestCaseType = ALLOWED_LANGUAGES[language.lower().strip()]
        self.assignment_name = cfg['ASSIGNMENT_NAME']
        source = cfg['SOURCE_FILE_NAME']
        if source == "AUTO":
            source = DEFAULT_SOURCE_FILE_STEM + self.TestCaseType.source_suffix
            self.auto_source_file_name_enabled = True
        else:
            self.auto_source_file_name_enabled = False
        self.lower_source_filename = cfg.getboolean('LOWER_SOURCE_FILENAME')
        if self.lower_source_filename:
            source = source.lower()
        self.source_file_name = source
        self.filters = [ALLOWED_FILTERS[f.strip()] for f in cfg['FILTERS'].split(",") if f]

    def _read_config(self) -> configparser.SectionProxy:
        default_parser = configparser.ConfigParser()
        default_parser.read(PATH_TO_DEFAULT_CONFIG)
        user_parser = configparser.ConfigParser()
        user_parser.read_dict(default_parser)
        user_parser.read(self.tests_dir / "config.ini")
        return user_parser['CONFIG']

    def _figure_out_testcase_type(self):
        for testcase in self.testcases_dir.iterdir():
            testcase_type = get_testcase_type_by_suffix(testcase.suffix)
            if testcase_type is not None:
                return testcase_type
        raise FileNotFoundError(
            f"Couldn't discover a testcase with correct suffix in {self.testcases_dir}"
        )

    def _configure_logging(self):
        self.logger = logging.getLogger("Grader")
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(logging.StreamHandler(sys.stdout))
        self.logger.addHandler(logging.FileHandler(self.path_to_output_summary, mode="w"))

    def _gather_testcases(self) -> List[testcases.TestCase]:
        tests = []
        for test in (self.testcases_dir).iterdir():
            if not (test.is_file() and self.TestCaseType.source_suffix in test.name):
                continue
            shutil.copy(test, self.temp_dir)
            tests.append(self.TestCaseType(
                self.temp_dir / test.name,
                self.tests_dir,
                self.timeout,
                self.filters,
                self.precompile_testcases
            ))
        return tests

    def _gather_submissions(self):
        submissions = []
        for submission in self.current_dir.iterdir():
            submission_name = submission.name
            if self.lower_source_filename:
                submission_name = submission_name.lower()
            if self.auto_source_file_name_enabled:
                if submission.suffix == self.TestCaseType.source_suffix:
                    submissions.append(submission)
            elif self.source_file_name in submission_name:
                submissions.append(submission)
        return submissions

    def _run_tests_on_submission(self, submission: Path):
        if self.generate_results:
            with open(self.results_dir / submission.name, "w") as f:
                grader_output = self._get_testcase_output(submission)
                f.write(format_grader_output(grader_output))
        else:
            grader_output = self._get_testcase_output(submission)
        return grader_output['student_score']

    def _get_testcase_output(self, submission) -> dict:
        """ Returns grading info as a dict """
        testcase_count = len(self.tests)
        if "_" in submission.name:
            student_name = submission.name[:submission.name.find("_")]
        else:
            student_name = submission.name
        self.logger.info(f"Grading {student_name}")
        try:
            precompiled_submission = self.TestCaseType.precompile_submission(
                submission, self.current_dir, self.source_file_name
            )
        except sh.ErrorReturnCode_1 as e:
            stderr = get_stderr(self.current_dir, e, "Failed to precompile")
            self.logger.info(stderr + f"\nResult: 0/{self.total_points_possible}\n")
            return {
                'assignment_name': self.assignment_name,
                'precompilation_error': stderr.replace('Failed to precompile', '')
            }
        total_testcase_score = 0
        testcase_results = []
        for test in self.tests:
            self.logger.info(f"Running '{test.name}'")
            testcase_score, message = test.run(precompiled_submission)
            self.logger.info(message)
            testcase_results.append((test.name, message))
            total_testcase_score += testcase_score
        student_score = total_testcase_score / testcase_count * self.total_score_to_100_ratio
        student_final_result = f"{round(student_score)}/{self.total_points_possible}"
        self.logger.info(f"Result: {student_final_result}\n")
        return {
            'assignment_name': self.assignment_name,
            'testcase_results': testcase_results,
            'formatted_student_score': student_final_result,
            'student_score': student_score
        }


def get_testcase_type_by_suffix(suffix: str):
    for testcase_type in ALLOWED_LANGUAGES.values():
        if testcase_type.source_suffix == suffix:
            return testcase_type


def format_grader_output(output: dict):
    """ Replace this function with anything else if you want the output
        to have a different style
    """
    s = f"{output['assignment_name']} Test Results\n\n"
    s += "%-40s%s" % ("TestCase", "Result")
    s += "\n================================================================"
    if "precompilation_error" in output:
        s += output['precompilation_error']
        return s
    for test_output in output['testcase_results']:
        s += "\n%-40s%s" % test_output
    s += "\n================================================================\n"
    s += "Result: " + output['formatted_student_score']
    s += KEY
    return s



import sys
import argparse
from pathlib import Path

from autograder.grader import Grader  # That's some awful naming
from autograder.util import print_results


def main(argv=None):
    """ Returns the average score of the students """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument('submission_path',
        type=Path, nargs="?", default=Path.cwd(),
        help='Path to directory that contains student submissions'
    )
    parser.add_argument('-g', '--generate_results',
        action="store_true",
        help='Generate results directory with a result file per student'
    )
    parser.add_argument('-p', '--print',
        type=float, nargs="?", default=None, const=100, metavar="min_score",
        help='Use after already graded to print assignments with score >= min_score'
    )
    parser.add_argument('--precompile_testcases',
        action="store_true",
        help='Precompile testcases to hide their files from student (Java support is minimal)'
    )
    args = parser.parse_args(argv)
    current_dir = (Path.cwd() / args.submission_path).resolve()
    if args.print is None:
        return Grader(
            current_dir,
            args.generate_results,
            precompile_testcases=args.precompile_testcases).run()
    else:
        print_results(current_dir, args.print)
        return -1


if __name__ == "__main__":
    main()
